"""See documentation on ``CFLPath``"""

from __future__ import annotations

import logging
import math
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from sqlalchemy.dialects.postgresql import ARRAY, array
from sqlalchemy.orm import aliased
from sqlalchemy.sql.expression import case, exists, func, insert, join, literal, select, true, union
from sqlalchemy.types import Boolean, String

from mate_common.models.cpg_types.llvm import Opcode
from mate_common.models.cpg_types.mate import EdgeKind, NodeKind
from mate_query.db import (
    BOT,
    AdvisoryLockBuildTransaction,
    Path,
    _make_edge_symbol_table,
    _make_transition_table,
    set_n_distinct,
)

if TYPE_CHECKING:
    from typing import ClassVar, Type

    from sqlalchemy.sql.expression import ClauseElement, Select

    from mate_query.db import Graph as CPG


logger = logging.getLogger(__name__)


class CFLPath(ABC, Path):
    """A Context-Free-Language path query.

    CFL queries find paths in the CPG such that concatenation of all the
    symbols/labels of the edges of the path form a word in some context-free
    language. Incremental detection of CFL-paths is done by a push-down
    automaton (PDA) encoded as a recursive SQL query.

    These 'queries' actually consist of a few distinct database operations:

    1. ``populate_edge_symbol_table`` constructs a table of symbols for (a relevant subset of) CPG edges. The mapping from edges to symbols is surjective and functional.
    2. ``populate_transition_table`` constructs a table of transitions for this PDA, where a transition consists of a functional mapping from a (source state, source stack, input symbol) tuple to a (new state, new stack) tuple.
    3. The recursive query that finds CFL-paths is defined in ``db.PathBuilder.build``.
    """

    cfl: ClassVar[bool] = True

    @classmethod
    @abstractmethod
    def table_name_prefix(cls) -> str:
        """Returns a table prefix to be used when creating edge symbol and transition tables.

        If multiple instantiations of a ``Path`` class return the same prefix, the resulting tables
        will be shared between instances, reducing the amount of pre-computation required.

        If different instances require distinct tables, ``table_name_prefix`` should return
        distinct prefixes.

        The name will be converted to lower case and truncated to 16 characters.
        """
        pass

    @classmethod
    def edge_symbol_table_name(cls, graph: CPG) -> str:
        # NB: Maximum length for SQL table name is apparently 63 chars, which
        # is why we truncate.
        return cls.table_name_prefix().lower()[:16] + "_" + graph.build

    @classmethod
    def transition_symbol_table_name(cls, graph: CPG) -> str:
        # NB: Maximum length for SQL table name is apparently 63 chars, which
        # is why we truncate.
        return (
            cls.table_name_prefix().lower()[:16]
            + ("_rev" if cls.is_reversed else "")
            + "_"
            + graph.build
        )

    @classmethod
    def _create_edge_symbol_table(cls, graph: CPG) -> bool:
        """This table only needs to be created once per CPG. Because it is not possible to
        atomically create the table from the selection (since ``CREATE IF NOT EXISTS ... AS ...`` is
        not atomic), this method requests a Postgres transaction-level advisory lock.

        If the table already exists, this method releases the lock immediately and returns
        ``False``. If this method creates the table for the first time, the lock is held until the
        next commit or rollback, and this method returns ``True``.
        """
        graph.session.execute(AdvisoryLockBuildTransaction(graph.build))
        cls.edge_symbol_table = _make_edge_symbol_table(
            cls.edge_symbol_table_name(graph), graph.BaseEdge.__table__
        )
        cls.edge_symbol_table.create(bind=graph.session.bind, checkfirst=True)
        if graph.session.query(cls.edge_symbol_table).first() is not None:
            graph.session.commit()  # free the lock
            return False
        return True

    @classmethod
    def _create_transition_table(cls, graph: CPG) -> bool:
        """This table only needs to be created once per CPG. Because it is not possible to
        atomically create the table from the selection (since ``CREATE IF NOT EXISTS ... AS ...`` is
        not atomic), this method requests a Postgres transaction-level advisory lock.

        If the table already exists, this method releases the lock immediately and returns
        ``False``. If this method creates the table for the first time, the lock is held until the
        next commit or rollback, and this method returns ``True``.
        """
        graph.session.execute(AdvisoryLockBuildTransaction(graph.build))
        cls.transition_table = _make_transition_table(cls.transition_symbol_table_name(graph))
        cls.transition_table.create(bind=graph.session.bind, checkfirst=True)
        if graph.session.query(cls.transition_table).first() is not None:
            graph.session.commit()  # free the lock
            return False
        return True

    @classmethod
    @abstractmethod
    def populate_transition_table(cls, graph: CPG) -> None:
        pass

    @classmethod
    @abstractmethod
    def populate_edge_symbol_table(cls, graph: CPG) -> None:
        pass

    @classmethod
    def update_transition_table_statistics(cls, graph: CPG) -> None:
        # Manually override postgres' n_distinct statistics for the transition table.
        #
        # We do this to avoid poor query planning during recursive CFL exploration by
        # providing better estimates for the three-way join:
        #              exploration <-> edge symbol <-> transition table
        # with the join conditions:
        #    base_select.c.target       == edge_symbol.table.c.source (or vice versa in reverse)
        #    base_select.c.state        == transition_table.c.old_state
        #    base_select.c.stack_top    == transition_table.c.old_stack
        #    edge_symbol_table.c.symbol == transition_table.c.input
        #
        # Because the base_select part of this join is materialized during the query,
        # the planner cannot use multi-variable statistics about the correlations between the
        # tables and falls back on estimating selectivity as the product of 1/n_distinct for
        # each referenced column. This is consistently a large underestimate, partially because
        # of the correlated values of old_stack and input in the transition table. This results
        # in very low selectivity estimates and poor plans.
        #
        # To approximate the multi-variable statistics, we will manually set n_distinct for
        # each of input, old_state, and old_stack such that their product is the number of
        # distinct entries in the table. In general, n_distinct(old_state) is very small and the
        # inaccuracy is due to the correlation of input and old_stack, so we will leave old_state
        # unchanged but set n_distinct(input) and n_distinct(old_stack) to
        # sqrt(n_distinct(entries) / n_distinct(old_state)).
        tablename = cls.transition_table.name  # type: ignore[union-attr]
        (entries, states) = graph.session.execute(
            f"SELECT COUNT(DISTINCT (input,old_stack,old_state)),"
            f"       COUNT(DISTINCT (old_state)) FROM {tablename}"
        ).fetchone()
        n_distinct = math.ceil(math.sqrt(entries / states))

        logger.debug(f"Updating statistics for {tablename}")
        logger.debug(f"{entries} distinct (input, old_stack, old_state) triples")
        logger.debug(f"{states} distinct (old_state) values")

        logger.debug(f"Setting (input, old_stack) n_distinct each to {n_distinct}")
        set_n_distinct(
            graph.session,
            tablename,
            set(["old_stack", "input"]),
            n_distinct,
        )

        logger.debug(f"Setting (old_state) n_distinct each to {states}")
        set_n_distinct(
            graph.session,
            tablename,
            set(["old_state"]),
            states,
        )

    @classmethod
    def update_edge_symbol_table_statistics(cls, graph: CPG) -> None:
        # Manually override postgres' n_distinct statistics for the edge_symbol_table
        #
        # We do this to avoid poor query planning during recursive CFL exploration by
        # providing better estimates for the three-way join:
        #              exploration <-> edge symbol <-> transition table
        # with the join conditions:
        #    base_select.c.target       == edge_symbol.table.c.source (or vice versa in reverse)
        #    base_select.c.state        == transition_table.c.old_state
        #    base_select.c.stack_top    == transition_table.c.old_stack
        #    edge_symbol_table.c.symbol == transition_table.c.input
        #
        # Because the base_select part of this join is materialized during the query,
        # the planner cannot use multi-variable statistics about the correlations between the
        # tables and falls back on estimating selectivity as the product of 1/n_distinct for
        # each referenced column. This is consistently a large underestimate, partially because
        # of the correlated values of old_stack and input in the transition table. This results
        # in very low selectivity estimates and poor plans.
        #
        # To approximate the multi-variable statistics, we will manually set n_distinct for
        # each of source, target, and symbol such that the product of n_distinct(symbol) and
        # either n_distinct(source) or n_distinct(target) is approximately the number of
        # distinct entries in the table.
        #
        # Because we expect a relatively small number of symbols per source / target and
        # want source and target to appear more selective to encourage joins on these first,
        # we will set n_distinct(source) and n_distinct(target) to their actual values but
        # set n_distinct(symbol) to n_distinct(entries) / max(n_distinct(source),n_distinct(target)).
        tablename = cls.edge_symbol_table.name  # type: ignore[union-attr]
        (entries, sources, targets, symbols) = graph.session.execute(
            f"SELECT COUNT(DISTINCT (source,target,symbol)),"
            f"       COUNT(DISTINCT (source)),"
            f"       COUNT(DISTINCT (target)),"
            f"       COUNT(DISTINCT (symbol)) FROM {tablename}"
        ).fetchone()

        new_symbols = math.ceil(entries / max(sources, targets))

        logger.debug(f"Updating statistics for {tablename}")
        logger.debug(f"{entries} distinct (source,target,symbol) triples")
        logger.debug(f"{sources} distinct (source) values")
        logger.debug(f"{targets} distinct (target) values")
        logger.debug(f"{symbols} distinct (symbol) values")

        logger.debug(f"Setting (source) n_distinct to {sources}")
        set_n_distinct(
            graph.session,
            tablename,
            set(["source"]),
            sources,
        )

        logger.debug(f"Setting (target) n_distinct to {targets}")
        set_n_distinct(
            graph.session,
            tablename,
            set(["target"]),
            targets,
        )

        logger.debug(f"Setting (symbol) n_distinct to {new_symbols}")
        set_n_distinct(
            graph.session,
            tablename,
            set(["symbol"]),
            new_symbols,
        )


def ConfigureCFGPath(*, allow_unmatched_returns: bool = False) -> Type[Path]:
    class CFGPath(CFLPath):
        """Inter-procedural control-flow graph (CFG) paths.

        The PDA encoded by this class has:

        * States:
            * ``INTRA``: The next step should follow intra-procedural control
              flow
            * ``INTER``: The next step should follow inter-procedural control
              flow

        * Edge symbols (variables appear ``<in angle brackets>``, tokens are
          separated by spaces for clarity):

            * Symbol:
              * ``step_instruction``
              * Edge kinds: ``EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION``
              * Condition: The target node is an instruction which is not a
                ``call``, ``invoke`` or ``ret``, **or** it is a ``call`` or
                invoke to a externally defined function (i.e., an LLVM
                declaration).

            * Symbol:
                * ``step_instruction_stop``
                * Edge kinds: ``EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION``
                * Condition: The target node is a ``call``, ``invoke`` or
                  ``ret`` instruction to a defined function.

            * Symbol:
                * ``call : <call instruction>``
                * Edge kinds: ``EdgeKind.CALL_TO_FUNCTION``
                * Condition: None

            * Symbol:
                * ``entry_block``
                * Edge kinds: ``EdgeKind.FUNCTION_TO_ENTRY_BLOCK``o
                * Condition: None

            * Symbol:
                * ``entry_instruction``
                * Edge kinds: ``EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION``
                * Condition: None

            * Symbol:
                * ``return_instruction``
                * Edge kinds: ``EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN``
                * Condition: None

            * Symbol:
                * ``return : <instruction>``
                * Edge kinds: ``EdgeKind.CALL_RETURN_TO_CALLER``
                * Condition: None

        * Transitions (variables appear ``<in angle brackets>``, ``push`` is
          the cons constructor for the stack type, ``bot`` is the bottom of the stack):
            * Transition:
                * Input state: ``INTRA``
                * Input stack: ``<s>``
                * Edge symbol: ``step_instruction``
                * New state: ``INTRA``
                * New stack: ``<s>``

            * Transition:
                * Input state: ``INTRA``
                * Input stack: ``<s>``
                * Edge symbol: ``step_instruction_stop``
                * New state: ``INTER``
                * New stack: ``<s>``

            * Transition:
                * Input state: ``INTER``
                * Input stack: ``<s>``
                * Edge symbol: ``call : <call instruction>``
                * New state: ``INTRA``
                * New stack: ``push(<call instruction>, <s>)``

            * Transition:
                * Input state: ``INTRA``
                * Input stack: ``<s>``
                * Edge symbol: ``entry_block``
                * New state: ``INTRA``
                * New stack: ``<s>``

            * Transition:
                * Input state: ``INTRA``
                * Input stack: ``<s>``
                * Edge symbol: ``entry_instruction``
                * New state: ``INTRA``
                * New stack: ``<s>``

            * Transition:
                * Input state: ``INTER``
                * Input stack: ``<s>``
                * Edge symbol: ``return_instruction``
                * New state: ``INTRA``
                * New stack: ``<s>``

            * Transition:
                * Input state: ``INTRA``
                * Input stack: ``push(<call instruction>, <s>)``
                * Edge symbol: ``return : <call instruction>``
                * New state: ``INTRA``
                * New stack: ``<s>``

            * Transition (only enabled if ``allow_unmatched_returns`` is ``True``):
                * Input state: ``INTRA``
                * Input stack: ``push(<s>)``
                * Edge symbol: ``return : <call instruction>``
                * New state: ``INTRA``
                * New stack: ``push(bot, <s>)``
        """

        @classmethod
        def table_name_prefix(cls) -> str:
            # The tables can be re-used as long as we're not stepping over any functions, but
            # the tables are different depending on whether unmatched returns are allowed.
            if not cls.stepped_over_functions:
                return "cfg_unmatched" if allow_unmatched_returns else "cfg"
            else:
                # Use the randomly generated Path name
                return cls.__name__

        @classmethod
        def initial_state(cls) -> ClauseElement:
            return literal("INTRA").cast(String)

        @classmethod
        def initial_stack(cls) -> ClauseElement:
            return array([BOT]).cast(ARRAY(String))

        @classmethod
        def populate_edge_symbol_table(cls, graph: CPG) -> None:
            """Preprocess the graph's edge table and give each edge a symbol to be used during
            during CFL-based path extension."""
            if not cls._create_edge_symbol_table(graph):
                return

            current_edge = graph.Edge
            current_node = aliased(graph.Node)
            outgoing_edge = aliased(graph.Edge)
            outgoing_node = aliased(graph.Node)

            # NOTE(lb): Any update to this query should be reflected in the
            # class docstring.
            edge_symbol_query = (
                select(
                    [
                        current_edge.uuid,
                        current_edge.source,
                        current_edge.target,
                        case(
                            [
                                (
                                    (
                                        current_edge.kind
                                        == EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION
                                    )
                                    & (
                                        ~(
                                            current_node.attributes["opcode"].astext.in_(
                                                [
                                                    Opcode.CALL.value,
                                                    Opcode.INVOKE.value,
                                                    Opcode.RET.value,
                                                ]
                                            )
                                        )
                                        | exists(
                                            select([outgoing_node]).select_from(
                                                join(
                                                    outgoing_edge,
                                                    outgoing_node,
                                                    (
                                                        (current_node.uuid == outgoing_edge.source)
                                                        & (
                                                            outgoing_edge.target
                                                            == outgoing_node.uuid
                                                        )
                                                        & (
                                                            outgoing_edge.kind
                                                            == EdgeKind.CALL_TO_FUNCTION
                                                        )
                                                        & (
                                                            # Step over (i.e. do not step into) any function which is
                                                            # just a declaration, or any function that has been
                                                            # explicitly marked as to be stepped over. Since this is a
                                                            # CASE statement with first-match-wins, these do not need to
                                                            # be excluded in the next case (step_instruction_stop).
                                                            (
                                                                outgoing_node.attributes[
                                                                    "is_declaration"
                                                                ].cast(Boolean)
                                                            )
                                                            | (
                                                                outgoing_node.attributes[
                                                                    "name"
                                                                ].astext.in_(
                                                                    cls.stepped_over_functions
                                                                )
                                                            )
                                                        )
                                                    ),
                                                )
                                            )
                                        )
                                    ),
                                    literal("step_instruction"),
                                ),
                                (
                                    (
                                        current_edge.kind
                                        == EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION
                                    )
                                    & (
                                        (
                                            current_node.attributes["opcode"].astext.in_(
                                                [
                                                    Opcode.CALL.value,
                                                    Opcode.INVOKE.value,
                                                    Opcode.RET.value,
                                                ]
                                            )
                                        )
                                    ),
                                    literal("step_instruction_stop"),
                                ),
                                (
                                    (current_edge.kind == EdgeKind.CALL_TO_FUNCTION),
                                    func.concat(
                                        literal("call"),
                                        ":",
                                        current_edge.source,
                                    ),
                                ),
                                (
                                    (current_edge.kind == EdgeKind.FUNCTION_TO_ENTRY_BLOCK),
                                    literal("entry_block"),
                                ),
                                (
                                    (current_edge.kind == EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION),
                                    literal("entry_instruction"),
                                ),
                                (
                                    (
                                        current_edge.kind
                                        == EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN
                                    ),
                                    literal("return_instruction"),
                                ),
                                (
                                    (current_edge.kind == EdgeKind.CALL_RETURN_TO_CALLER),
                                    func.concat(
                                        literal("return"),
                                        ":",
                                        current_edge.target,
                                    ),
                                ),
                            ],
                        ).label("symbol"),
                    ]
                )
                .select_from(
                    join(current_edge, current_node, current_edge.target == current_node.uuid)
                )
                .alias()
            )

            graph.session.execute(
                insert(cls.edge_symbol_table).from_select(
                    ["uuid", "source", "target", "symbol"],
                    select(
                        [
                            edge_symbol_query.c.uuid,
                            edge_symbol_query.c.source,
                            edge_symbol_query.c.target,
                            edge_symbol_query.c.symbol,
                        ]
                    ).where(edge_symbol_query.c.symbol != None),
                )
            )
            cls.update_edge_symbol_table_statistics(graph)
            graph.session.commit()

        @classmethod
        def populate_transition_table(cls, graph: CPG) -> None:
            if not cls._create_transition_table(graph):
                return

            all_stack_symbols = union(
                select([graph.Edge.source.label("stack_symbol")]).where(
                    graph.Edge.kind == EdgeKind.CALL_TO_FUNCTION
                ),
                select([literal(BOT).label("stack_symbol")]),
            ).alias()

            # NOTE(lb): Any update to these rules should be reflected in the
            # class docstring.
            rules = {
                "step_instruction": select(
                    [
                        literal("INTRA"),
                        all_stack_symbols.c.stack_symbol,
                        literal("step_instruction"),
                        literal("INTRA"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
                "step_instruction_stop": select(
                    [
                        literal("INTRA"),
                        all_stack_symbols.c.stack_symbol,
                        literal("step_instruction_stop"),
                        literal("INTER"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
                "call": select(
                    [
                        literal("INTER"),
                        all_stack_symbols.c.stack_symbol,
                        func.concat(
                            literal("call"),
                            ":",
                            graph.Edge.source,
                        ),
                        literal("INTRA"),
                        array(
                            [
                                graph.Edge.source,
                                all_stack_symbols.c.stack_symbol,
                            ]
                        ).cast(ARRAY(String)),
                    ],
                ).select_from(
                    join(
                        all_stack_symbols,
                        graph.Edge,
                        (graph.Edge.kind == EdgeKind.CALL_TO_FUNCTION),
                    )
                ),
                "entry_block": select(
                    [
                        literal("INTRA"),
                        all_stack_symbols.c.stack_symbol,
                        literal("entry_block"),
                        literal("INTRA"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ]
                ).select_from(all_stack_symbols),
                "entry_instruction": select(
                    [
                        literal("INTRA"),
                        all_stack_symbols.c.stack_symbol,
                        literal("entry_instruction"),
                        literal("INTRA"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ]
                ).select_from(all_stack_symbols),
                "return_instruction": select(
                    [
                        literal("INTER"),
                        all_stack_symbols.c.stack_symbol,
                        literal("return_instruction"),
                        literal("INTRA"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ]
                ).select_from(all_stack_symbols),
                "return": select(
                    [
                        literal("INTRA"),
                        all_stack_symbols.c.stack_symbol,
                        func.concat(literal("return"), ":", all_stack_symbols.c.stack_symbol),
                        literal("INTRA"),
                        array([]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
            }

            if allow_unmatched_returns:
                rules["unmatched_return"] = select(
                    [
                        literal("INTRA"),
                        literal(BOT),
                        func.concat(literal("return"), ":", all_stack_symbols.c.stack_symbol),
                        literal("INTRA"),
                        array([BOT]).cast(ARRAY(String)),
                    ]
                ).select_from(all_stack_symbols)

            graph.session.execute(
                insert(cls.transition_table).from_select(
                    ["old_state", "old_stack", "input", "new_state", "new_stack"],
                    union(*[rules[i] for i in rules]),
                )
            )
            cls.update_transition_table_statistics(graph)
            graph.session.commit()

    return CFGPath


ForwardCFGPath = ConfigureCFGPath(allow_unmatched_returns=False)
"""Inter-procedural control-flow graph (CFG) paths, in the forwards direction.

Matches calls and returns up to unbounded depth, rather than using
context-sensitivity information from the pointer analysis.

When encountering a return when the call stack is empty, can return to any
callsite, which is problematic for performance and precision (in contract to
context-sensitive variants below which use points-to analysis information).
"""

UnmatchedForwardCFGPath = ConfigureCFGPath(allow_unmatched_returns=True)
"""Like ``ForwardCFGPath`` but doesn't match calls and returns.
"""


class ContextSensitiveCFLPath(CFLPath):
    @classmethod
    def select_callsite_stack_symbols(self, graph: CPG) -> Select:
        """Construct partial stack symbols of the form:

          <call instruction> : <caller context> --> <callee context>

        With columns:

        * ``stack_symbol``: The overall symbol
        * ``callsite``: caller instruction
        * ``caller``: calling function
        * ``callee``: called function
        * ``caller_context``
        * ``caller_context``

        This select is used as a variable in complex rules in PDAs, where the
        above partial stack symbol gets concatenated onto the tokens ``call:``
        or ``return:``, e.g.

        .. code-block::

             call : <call instruction> : <caller context> --> <callee context>

        This overall symbol gets pushed to or popped from the PDA stack during
        transitions.
        """
        block_edge = aliased(graph.Edge)
        function_edge = aliased(graph.Edge)
        return (
            select(
                [
                    func.concat(
                        graph.Edge.source,
                        ":",
                        graph.Edge.attributes["caller_context"].astext,
                        "-->",
                        graph.Edge.attributes["callee_context"].astext,
                    ).label("stack_symbol"),
                    graph.Edge.source.label("callsite"),
                    graph.Edge.attributes["caller_context"].astext.label("caller_context"),
                    function_edge.target.label("caller"),
                    graph.Edge.attributes["callee_context"].astext.label("callee_context"),
                    graph.Edge.target.label("callee"),
                ]
            ).select_from(
                join(
                    join(
                        graph.Edge,
                        block_edge,
                        (graph.Edge.kind == EdgeKind.CALL_TO_FUNCTION)
                        & (block_edge.kind == EdgeKind.INSTRUCTION_TO_PARENT_BLOCK)
                        & (graph.Edge.source == block_edge.source),
                    ),
                    function_edge,
                    (function_edge.kind == EdgeKind.BLOCK_TO_PARENT_FUNCTION)
                    & (block_edge.target == function_edge.source),
                )
            )
        ).alias()

    @classmethod
    def select_context_stack_symbols(self, graph: CPG) -> Select:
        """Construct stack symbols of the form ``<context>``.

        with columns:

        * ``stack_symbol``: The context
        * ``function``: The function

        where there exists some callgraph edge out of (resp., into)
        ``<function>`` with caller (resp., callee) context ``<context>``.

        This select is used as a variable in complex rules in PDAs.
        """
        return union(
            select(
                [
                    graph.Edge.attributes["caller_context"].astext.label("stack_symbol"),
                    graph.Edge.source.label("function"),
                ]
            )
            .where(graph.Edge.kind == EdgeKind.CALLGRAPH)
            .select_from(graph.Edge),
            select(
                [
                    graph.Edge.attributes["callee_context"].astext.label("stack_symbol"),
                    graph.Edge.target.label("function"),
                ]
            )
            .where(graph.Edge.kind == EdgeKind.CALLGRAPH)
            .select_from(graph.Edge),
        ).alias()


class CSCFGPath(ContextSensitiveCFLPath):
    """Inter-procedural context-sensitive (CS) control-flow graph (CFG) paths.

    These use context-sensitive information from the points-to analysis, and do
    call-return matching. If an unmatched return is encountered, then the PDA
    returns to any feasible control-flow context of the caller (as determined
    by the MATE pointer analysis).

    In the PDA for this class, "context" usually refers to a k-limited (where k
    is usually relatively low, like 1 or 2) list of callsites (stack frames).
    The PDA encoded by this class has:

    * States:
        * ``INTRA``: The next step should follow intra-procedural control
            flow
        * ``INTER``: The next step should follow inter-procedural control
            flow

    * Edge symbols (variables appear ``<in angle brackets>``, tokens are
      separated by spaces for clarity):

        * Symbol:
            * ``step_instruction``
            * Edge kinds: ``EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION``)
            * Condition: The target node is an instruction which is not a ``call``,
              ``invoke`` or ``ret``, **or** it is a ``call`` or invoke to a
              externally defined function (i.e., an LLVM declaration).

        * Symbol:
            * ``step_instruction_stop``
            * Edge kinds: ``EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION``
            * Condition: The target node is a ``call``, ``invoke`` or ``ret``
              instruction to a defined function.

        * Symbol:
            * ``call : <call instruction> : <caller context> --> <callee context>``
            * Edge kinds: ``EdgeKind.CALL_TO_FUNCTION``
            * Condition: The target node is a ``Function`` called by the source
              (instruction) node in (control-flow) context ``<caller
              context>``, resulting in context ``<callee context>``.

        * Symbol:
            * ``entry_block``
            * Edge kinds: ``EdgeKind.FUNCTION_TO_ENTRY_BLOCK``
            * Condition: None

        * Symbol:
            * ``entry_instruction``
            * Edge kinds: ``EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION``
            * Condition: None

        * Symbol:
            * ``return_instruction``
            * Edge kinds: ``EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN``
            * Condition: None

        * Symbol:
            * ``return : <instruction> : <caller context> --> <callee context>``
            * Edge kinds: ``EdgeKind.CALL_RETURN_TO_CALLER``
            * Condition: The target is the ``call`` or ``invoke`` instruction
              for the call to the function that the control flow is returning
              from. The call occurred in context ``<caller context>``,
              resulting in context ``<callee context>``.

    * Stack symbol templates/formats:
       * ``call : <call instruction> : <caller context> --> <callee context>``
       * ``<context>``

    * Transitions (variables appear ``<in angle brackets>``, ``push`` is the
      cons constructor for the stack type, ``bot`` is the bottom of the stack):

        * Transition:
            * Input state: ``INTRA``
            * Input stack: ``<s>``
            * Edge symbol: ``step_instruction``
            * New state: ``INTRA``
            * New stack: ``<s>``
        * Transition:
            * Input state: ``INTRA``
            * Input stack: ``<s>``
            * Edge symbol: ``step_instruction_stop``
            * New state: ``INTER``
            * New stack: ``<s>``
        * Transition:
            * Description: TODO(lb)
            * Input state: ``INTER``
            * Input stack: ``push(call : <call instruction 0> : <caller context 0> --> <callee context 0>)``
            * Edge symbol: ``call : <call instruction> : <caller context> --> <callee context>``
            * New state: ``INTRA``
            * New stack: ``push(call : <call instruction> : <caller context> --> <callee context>, push(call : <call instruction 0> : <caller context 0> : <callee context 0>))``
            * Where: ``<callee context 0> = <caller context>`` and the callee
              function of the existing stack frame is the caller context of the
              new stack frame.
            * TODO(lb): how does that relate to ``<call instruction>``?
        * Transition:
            * Description: Call with empty stack
            * Input state: ``INTER``
            * Input stack: ``bot``
            * Edge symbol: ``call : <call instruction> : <caller context> --> <callee context>``
            * New state: ``INTRA``
            * New stack: ``push(<call instruction> : <caller context> --> <callee context>, bot)``
        * Transition:
            * Description: TODO(lb)
            * Input state: ``INTER``
            * Input stack: ``push(<caller context>, <s>)``
            * Edge symbol: ``call : <call instruction> : <caller context> --> <callee context>``
            * New state: ``INTRA``
            * Input stack: ``push(call : <call instruction> : <caller context> --> <callee context>, push(<caller context>, <s>))``
            * Where: ``<call instruction>`` may call ``<function>``
        * Transition:
            * Input state: ``INTRA``
            * Input stack: ``<s>``
            * Edge symbol: ``entry_block``
            * New state: ``INTRA``
            * New stack: ``<s>``
        * Transition:
            * Input state: ``INTRA``
            * Input stack: ``<s>``
            * Edge symbol: ``entry_instruction``
            * New state: ``INTRA``
            * New stack: ``<s>``
        * Transition:
            * Input state: ``INTER``
            * Input stack: ``<s>``
            * Edge symbol: ``return_instruction``
            * New state: ``INTRA``
            * New stack: ``<s>``
        * Transition:
            * Description: Return with a call on the stack
            * Input state: ``INTRA``
            * Input stack: ``push(call : <call instruction> : <caller context> --> <callee context>, <s>)``
            * Edge symbol: ``return : <call instruction> : <caller context> --> <callee context>``
            * New state: ``INTRA``
            * New stack: ``<s>``
        * Transition:
            * Description: Return with an empty stack
            * Input state: ``INTRA``
            * Input stack: ``bot``
            * Edge symbol: ``return : <call instruction> : <caller context> --> <callee context>``
            * New state: ``INTRA``
            * New stack: ``push(<caller context>, bot)``
        * Transition:
            * Input state: ``INTRA``
            * Input stack: ``push(<callee context>, <s>)``
            * Edge symbol: ``return : <call instruction> : <caller context> : <callee context>``
            * New state: ``INTRA``
            * New stack: ``push(<caller context>, <s>)``
    """

    @classmethod
    def table_name_prefix(cls) -> str:
        # The tables can be re-used as long as we're not stepping over any functions
        if not cls.stepped_over_functions:
            return "cscfg"
        else:
            # Use the randomly generated Path name
            return cls.__name__

    @classmethod
    def initial_state(cls) -> ClauseElement:
        return literal("INTRA").cast(String)

    @classmethod
    def initial_stack(cls) -> ClauseElement:
        return array([BOT]).cast(ARRAY(String))

    @classmethod
    def populate_edge_symbol_table(cls, graph: CPG) -> None:
        """Preprocess the graph's edge table and give each edge a symbol to be used during during
        CFL-based path extension."""
        if not cls._create_edge_symbol_table(graph):
            return

        current_edge = graph.Edge
        current_node = aliased(graph.Node)
        outgoing_edge = aliased(graph.Edge)
        outgoing_node = aliased(graph.Node)

        # NOTE(lb): Any update to this query should be reflected in the class
        # docstring.
        edge_symbol_query = (
            select(
                [
                    current_edge.uuid,
                    current_edge.source,
                    current_edge.target,
                    case(
                        [
                            (
                                (current_edge.kind == EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION)
                                & (
                                    ~(
                                        current_node.attributes["opcode"].astext.in_(
                                            [
                                                Opcode.CALL.value,
                                                Opcode.INVOKE.value,
                                                Opcode.RET.value,
                                            ]
                                        )
                                    )
                                    | exists(
                                        select([outgoing_node]).select_from(
                                            join(
                                                outgoing_edge,
                                                outgoing_node,
                                                (
                                                    (current_node.uuid == outgoing_edge.source)
                                                    & (outgoing_edge.target == outgoing_node.uuid)
                                                    & (
                                                        outgoing_edge.kind
                                                        == EdgeKind.CALL_TO_FUNCTION
                                                    )
                                                    & (
                                                        # Step over (i.e. do not step into) any function which is
                                                        # just a declaration, or any function that has been
                                                        # explicitly marked as to be stepped over. Since this is a
                                                        # CASE statement with first-match-wins, these do not need to
                                                        # be excluded in the next case (step_instruction_stop).
                                                        (
                                                            outgoing_node.attributes[
                                                                "is_declaration"
                                                            ].cast(Boolean)
                                                        )
                                                        | (
                                                            outgoing_node.attributes[
                                                                "name"
                                                            ].astext.in_(cls.stepped_over_functions)
                                                        )
                                                    )
                                                ),
                                            )
                                        )
                                    )
                                ),
                                literal("step_instruction"),
                            ),
                            (
                                (current_edge.kind == EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION)
                                & (
                                    (
                                        current_node.attributes["opcode"].astext.in_(
                                            [
                                                Opcode.CALL.value,
                                                Opcode.INVOKE.value,
                                                Opcode.RET.value,
                                            ]
                                        )
                                    )
                                ),
                                literal("step_instruction_stop"),
                            ),
                            (
                                (current_edge.kind == EdgeKind.CALL_TO_FUNCTION),
                                func.concat(
                                    literal("call"),
                                    ":",
                                    current_edge.source,
                                    ":",
                                    current_edge.attributes["caller_context"].astext,
                                    "-->",
                                    current_edge.attributes["callee_context"].astext,
                                ),
                            ),
                            (
                                (current_edge.kind == EdgeKind.FUNCTION_TO_ENTRY_BLOCK),
                                literal("entry_block"),
                            ),
                            (
                                (current_edge.kind == EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION),
                                literal("entry_instruction"),
                            ),
                            (
                                (current_edge.kind == EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN),
                                literal("return_instruction"),
                            ),
                            (
                                (current_edge.kind == EdgeKind.CALL_RETURN_TO_CALLER),
                                func.concat(
                                    literal("return"),
                                    ":",
                                    current_edge.target,
                                    ":",
                                    current_edge.attributes["caller_context"].astext,
                                    "-->",
                                    current_edge.attributes["callee_context"].astext,
                                ),
                            ),
                        ],
                    ).label("symbol"),
                ]
            )
            .select_from(join(current_edge, current_node, current_edge.target == current_node.uuid))
            .alias()
        )

        graph.session.execute(
            insert(cls.edge_symbol_table).from_select(
                ["uuid", "source", "target", "symbol"],
                select(
                    [
                        edge_symbol_query.c.uuid,
                        edge_symbol_query.c.source,
                        edge_symbol_query.c.target,
                        edge_symbol_query.c.symbol,
                    ]
                ).where(edge_symbol_query.c.symbol != None),
            )
        )
        cls.update_edge_symbol_table_statistics(graph)
        graph.session.commit()

    @classmethod
    def populate_transition_table(cls, graph: CPG) -> None:
        if not cls._create_transition_table(graph):
            return

        callsite_stack_symbols = cls.select_callsite_stack_symbols(graph)
        callsite_stack_symbols2 = aliased(callsite_stack_symbols)
        context_stack_symbols = cls.select_context_stack_symbols(graph)

        all_stack_symbols = union(
            select([callsite_stack_symbols.c.stack_symbol]).select_from(callsite_stack_symbols),
            select([context_stack_symbols.c.stack_symbol]).select_from(context_stack_symbols),
            select([literal(BOT).label("stack_symbol")]),
        ).alias()

        # NOTE(lb): Any update to these rules should be reflected in the class
        # docstring.
        rules = {
            "step_instruction": select(
                [
                    literal("INTRA"),
                    all_stack_symbols.c.stack_symbol,
                    literal("step_instruction"),
                    literal("INTRA"),
                    array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                ],
            ).select_from(all_stack_symbols),
            "step_instruction_stop": select(
                [
                    literal("INTRA"),
                    all_stack_symbols.c.stack_symbol,
                    literal("step_instruction_stop"),
                    literal("INTER"),
                    array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                ],
            ).select_from(all_stack_symbols),
            "call": select(
                [
                    literal("INTER"),
                    callsite_stack_symbols2.c.stack_symbol,
                    func.concat(
                        literal("call"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array(
                        [
                            callsite_stack_symbols.c.stack_symbol,
                            callsite_stack_symbols2.c.stack_symbol,
                        ]
                    ).cast(ARRAY(String)),
                ],
            ).select_from(
                join(
                    callsite_stack_symbols,
                    callsite_stack_symbols2,
                    (callsite_stack_symbols2.c.callee == callsite_stack_symbols.c.caller)
                    & (
                        callsite_stack_symbols2.c.callee_context
                        == callsite_stack_symbols.c.caller_context
                    ),
                )
            ),
            "empty_call": select(
                [
                    literal("INTER"),
                    literal(BOT),
                    func.concat(
                        literal("call"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array([callsite_stack_symbols.c.stack_symbol, BOT]).cast(ARRAY(String)),
                ],
            ).select_from(callsite_stack_symbols),
            "context_call": select(
                [
                    literal("INTER"),
                    context_stack_symbols.c.stack_symbol,
                    func.concat(
                        literal("call"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array(
                        [
                            callsite_stack_symbols.c.stack_symbol,
                            context_stack_symbols.c.stack_symbol,
                        ]
                    ).cast(ARRAY(String)),
                ],
            ).select_from(
                join(
                    context_stack_symbols,
                    callsite_stack_symbols,
                    (
                        context_stack_symbols.c.stack_symbol
                        == callsite_stack_symbols.c.caller_context
                    )
                    & (context_stack_symbols.c.function == callsite_stack_symbols.c.caller),
                )
            ),
            "entry_block": select(
                [
                    literal("INTRA"),
                    all_stack_symbols.c.stack_symbol,
                    literal("entry_block"),
                    literal("INTRA"),
                    array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                ]
            ).select_from(all_stack_symbols),
            "entry_instruction": select(
                [
                    literal("INTRA"),
                    all_stack_symbols.c.stack_symbol,
                    literal("entry_instruction"),
                    literal("INTRA"),
                    array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                ]
            ).select_from(all_stack_symbols),
            "return_instruction": select(
                [
                    literal("INTER"),
                    all_stack_symbols.c.stack_symbol,
                    literal("return_instruction"),
                    literal("INTRA"),
                    array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                ]
            ).select_from(all_stack_symbols),
            "return": select(
                [
                    literal("INTRA"),
                    callsite_stack_symbols.c.stack_symbol,
                    func.concat(
                        literal("return"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array([]).cast(ARRAY(String)),
                ],
            ).select_from(callsite_stack_symbols),
            "empty_return": select(
                [
                    literal("INTRA"),
                    literal(BOT),
                    func.concat(
                        literal("return"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array([callsite_stack_symbols.c.caller_context, literal(BOT)]).cast(
                        ARRAY(String)
                    ),
                ],
            ).select_from(callsite_stack_symbols),
            "context_return": select(
                [
                    literal("INTRA"),
                    callsite_stack_symbols.c.callee_context,
                    func.concat(
                        literal("return"),
                        ":",
                        callsite_stack_symbols.c.stack_symbol,
                    ),
                    literal("INTRA"),
                    array([callsite_stack_symbols.c.caller_context]).cast(ARRAY(String)),
                ],
            ).select_from(callsite_stack_symbols),
        }

        graph.session.execute(
            insert(cls.transition_table).from_select(
                ["old_state", "old_stack", "input", "new_state", "new_stack"],
                union(*[rules[i] for i in rules]),
            )
        )
        cls.update_transition_table_statistics(graph)
        graph.session.commit()


def ConfigureDFGPath(*, thin: bool = True) -> Type[Path]:
    class CSDataflowPath(ContextSensitiveCFLPath):
        """Thin dataflow paths that use context-sensitivity (CS) from the points-to analysis.

        The PDA encoded by this class has:

        * States:
            * ``DATAFLOW``
        * Edge symbols (variables appear ``<in angle brackets>``, tokens are
          separated by spaces for clarity):

            * Symbol:
                * ``value_definition_to_use``
                * Edge kinds: ``EdgeKind.VALUE_DEFINITION_TO_USE``
                * Condition: The target node is an instruction which is not a
                  ``call``, ``invoke`` or ``load``; **and** the use is not as the
                  first operand of a ``store``.

            * Symbol:
                * ``dataflow_from_memory : <context>``
                * Edge kinds: ``EdgeKind.DATAFLOW_SIGNATURE``,
                  ``EdgeKind.DIRECT_DATAFLOW_SIGNATURE``, plus
                  ``EdgeKind.INDIRECT_DATAFLOW_SIGNATURE`` if ``thin`` is
                  ``False``.
                * Condition: The target node is a dataflow, input, or output
                  signature, and the source node is a memory location.

            * Symbol:
                * ``dataflow_from_value : <context>``
                * Edge kinds: ``EdgeKind.DATAFLOW_SIGNATURE``,
                  ``EdgeKind.DIRECT_DATAFLOW_SIGNATURE``, plus
                  ``EdgeKind.INDIRECT_DATAFLOW_SIGNATURE`` if ``thin`` is
                  ``False``.
                * Condition: The target node is a dataflow, input, or output
                  signature, and the source node is *not* a memory location (i.e.,
                  it is an LLVM value).

            * Symbol:
                * ``dataflow_from_value : <context>``
                * Edge kinds: ``EdgeKind.DATAFLOW_SIGNATURE``,
                  ``EdgeKind.DIRECT_DATAFLOW_SIGNATURE``, plus
                  ``EdgeKind.INDIRECT_DATAFLOW_SIGNATURE`` if ``thin`` is
                  ``False``.
                * Condition: The target node is a dataflow, input, or output
                  signature, and the source node is *not* a memory location (i.e.,
                  it is an LLVM value).

            * Symbol:
                * ``dataflow_to_memory : <context>``
                * Edge kinds: ``EdgeKind.DATAFLOW_SIGNATURE``,
                  ``EdgeKind.DIRECT_DATAFLOW_SIGNATURE``, plus
                  ``EdgeKind.INDIRECT_DATAFLOW_SIGNATURE`` if ``thin`` is
                  ``False``.
                * Condition: The target node is a memory location.

            * Symbol:
                * ``dataflow_to_value : <context>``
                * Edge kinds: ``EdgeKind.DATAFLOW_SIGNATURE``,
                  ``EdgeKind.DIRECT_DATAFLOW_SIGNATURE``, plus
                  ``EdgeKind.INDIRECT_DATAFLOW_SIGNATURE`` if ``thin`` is
                  ``False``.
                * Condition: The target node is *not* a memory location (i.e., it
                  is an LLVM value).

            * Symbol:
                * ``load : <context>``
                * Edge kinds: ``EdgeKind.LOAD_MEMORY``
                * Condition: None

            * Symbol:
                * ``store : <context>``
                * Edge kinds: ``EdgeKind.STORE_MEMORY``
                * Condition: None

            * Symbol:
                * ``operand_to_param_binding``
                * Edge kinds: ``EdgeKind.OPERAND_TO_PARAM_BINDING``
                * Condition: None

            * Symbol:
                * ``call : <call instruction> : <caller context> --> <callee context>``
                * Edge kinds: ``EdgeKind.PARAM_BINDING_TO_ARG``
                * Condition: The parameter was bound by this call in this calling
                  context

            * Symbol:
                * ``return_value_to_call_return``
                * Edge kinds: ``EdgeKind.RETURN_VALUE_TO_CALL_RETURN``
                * Condition: None

            * Symbol:
                * ``return : <call instruction> : <caller context> --> <callee context>``
                * Edge kinds: ``EdgeKind.CALL_RETURN_TO_CALLER``
                * Condition: This return is from a (potential) call in context
                  ``<caller context>``, resulting in callee context ``<callee
                  context>``.
        * Stack symbol templates/formats:
        * ``<call instruction> : <caller context> --> <callee context>``
        * ``<context>``
        * Transitions (variables appear ``<in angle brackets>``, ``push`` is the
          cons constructor for the stack type, ``bot`` is the bottom of the stack,
          all input and output states are ``DATAFLOW``):
            * Transition:
                * Input stack: ``<s>``
                * Edge symbol: ``value_definition_to_use``
                * New stack: ``<s>``
            * Transition:
                * Input stack: ``<s>``
                * Edge symbol: ``operand_to_param_binding``
                * New stack: ``<s>``
            * Transition:
                * Input stack: ``<s>``
                * Edge symbol: ``return_value_to_call_return``
                * New stack: ``<s>``
            * Transition:
                * Description: TODO(lb)
                * Input stack: ``push(call : <call instruction 0> : <caller context 0> --> <callee context 0>)``
                * Edge symbol: ``call : <call instruction> : <caller context> --> <callee context>``
                * New stack: ``push(call : <call instruction> : <caller context> --> <callee context>, push(call : <call instruction 0> : <caller context 0> : <callee context 0>))``
                * Where: ``<callee context 0> = <caller context>`` and the callee
                  function of the existing stack frame is the caller context of the
                  new stack frame.
                * TODO(lb): how does that relate to ``<call instruction>``?
            * Transition:
                * Input stack: ``push(<caller context>, <s>)``
                * Edge symbol: ``call : <call instruction> : <caller context> --> <callee context>``
                * New stack: ``push(<call instruction> : <caller context> --> <callee context>, push(<context>, bot))``
            * Transition:
                * Input stack: ``push(call : <call instruction> : <caller context> --> <callee context>, <s>)``
                * Edge symbol: ``return : <call instruction> : <caller context> --> <callee context>``
                * New stack: ``<s>``
            * Transition:
                * Description: Return with an empty stack
                * Input stack: ``bot``
                * Edge symbol: ``return : <call instruction> : <caller context> --> <callee context>``
                * New stack: ``push(<caller context>, bot)``
            * Transition:
                * Input stack: ``push(<callee context>, <s>)``
                * Edge symbol: ``return : <call instruction> : <caller context> : <callee context>``
                * New stack: ``push(<caller context>, <s>)``
            * Transition:
                * Input stack: ``bot``
                * Edge symbol: ``load : <context>``
                * New stack: ``push(<context>, <s>)``
            * Transition:
                * Input stack: ``push(<call instruction> : <caller context> --> <callee context>, <s>)``
                * Edge symbol: ``store : <callee context>``
                * New stack: ``bot``
            * Transition:
                * Input stack: ``bot``
                * Edge symbol: ``store : <context>``
                * New stack: ``bot``
            * Transition:
                * Input stack: ``push(<context>, <s>)``
                * Edge symbol: ``store : <context>``
                * New stack: ``bot``
            * Similar rules where ``dataflow_from_memory`` is to
              ``dataflow_to_memory`` as ``load`` is to ``store``...
            * Transition:
                * Input stack: ``push(<context>, <s>)``
                * Edge symbol: ``dataflow_from_value : <context>``
                * New stack: ``push(<context>, <s>)``
            * Transition:
                * Input stack: ``bot``
                * Edge symbol: ``dataflow_from_value : <context>``
                * New stack: ``push(<context>, bot)``
            * Transition:
                * Input stack: ``push(<call instruction> : <caller context> --> <callee context>, bot)``
                * Edge symbol: ``dataflow_from_value : <callee context>``
                * New stack: ``push(<call instruction> : <caller context> --> <callee context>, bot)``
            * Transition:
                * Input stack: ``push(<context>, <s>)``
                * Edge symbol: ``dataflow_to_value : <context>``
                * New stack: ``push(<context>, <s>)``
            * Transition:
                * Input stack: ``bot``
                * Edge symbol: ``dataflow_to_value : <context>``
                * New stack: ``push(<context>, bot)``
            * Transition:
                * Input stack: ``push(<call instruction> : <caller context> --> <callee context>, bot)``
                * Edge symbol: ``dataflow_to_value : <callee context>``
                * New stack: ``push(<call instruction> : <caller context> --> <callee context>, bot)``


        Transitions for a reversed automaton are dual in some sense...
        """

        @classmethod
        def table_name_prefix(cls) -> str:
            # The PDA tables are the same for all instances
            return "cs" + ("thin" if thin else "") + "dataflow"

        @classmethod
        def initial_state(cls) -> ClauseElement:
            return literal("DATAFLOW").cast(String)

        @classmethod
        def initial_stack(cls) -> ClauseElement:
            return array([BOT]).cast(ARRAY(String))

        @classmethod
        def populate_edge_symbol_table(cls, graph: CPG) -> None:
            """Preprocess the graph's edge table and give each edge a symbol to be used during
            during CFL-based path extension."""

            if not cls._create_edge_symbol_table(graph):
                return

            current_edge = graph.Edge
            source_node = aliased(graph.Node)
            target_node = aliased(graph.Node)
            current_call = aliased(graph.Node)
            current_call_edge = aliased(graph.Edge)

            # NOTE(lb): Any update to this query should be reflected in the class
            # docstring.
            edge_symbol_query = (
                select(
                    [
                        current_edge.uuid,
                        current_edge.source,
                        current_edge.target,
                        case(
                            [
                                (
                                    (current_edge.kind == EdgeKind.VALUE_DEFINITION_TO_USE)
                                    & (
                                        (
                                            (
                                                ~(
                                                    target_node.attributes["opcode"].astext.in_(
                                                        [
                                                            Opcode.CALL.value,
                                                            Opcode.INVOKE.value,
                                                            Opcode.LOAD.value,
                                                        ]
                                                    )
                                                )
                                            )
                                            if thin
                                            else true()
                                        )
                                    )
                                    & (
                                        (
                                            (
                                                (
                                                    target_node.attributes["opcode"].astext
                                                    != Opcode.STORE.value
                                                )
                                                | (
                                                    current_edge.attributes[
                                                        "operand_number"
                                                    ].as_integer()
                                                    != 1
                                                )
                                            )
                                            if thin
                                            else true()
                                        )
                                    ),
                                    literal("value_definition_to_use"),
                                ),
                                # Edges to dataflow signature nodes from memory
                                (
                                    (
                                        current_edge.kind.in_(
                                            [
                                                EdgeKind.DATAFLOW_SIGNATURE,
                                                EdgeKind.DIRECT_DATAFLOW_SIGNATURE,
                                            ]
                                            + (
                                                []
                                                if thin
                                                else [EdgeKind.INDIRECT_DATAFLOW_SIGNATURE]
                                            )
                                        )
                                        & target_node.kind.in_(
                                            [
                                                NodeKind.DATAFLOW_SIGNATURE,
                                                NodeKind.INPUT_SIGNATURE,
                                                NodeKind.OUTPUT_SIGNATURE,
                                            ]
                                        )
                                        & source_node.kind.in_(
                                            [
                                                NodeKind.MEMORY_LOCATION,
                                            ]
                                        )
                                    ),
                                    func.concat(
                                        literal("dataflow_from_memory"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                # Edges to dataflow signature nodes from values
                                (
                                    (
                                        current_edge.kind.in_(
                                            [
                                                EdgeKind.DATAFLOW_SIGNATURE,
                                                EdgeKind.DIRECT_DATAFLOW_SIGNATURE,
                                            ]
                                            + (
                                                []
                                                if thin
                                                else [EdgeKind.INDIRECT_DATAFLOW_SIGNATURE]
                                            )
                                        )
                                        & target_node.kind.in_(
                                            [
                                                NodeKind.DATAFLOW_SIGNATURE,
                                                NodeKind.INPUT_SIGNATURE,
                                                NodeKind.OUTPUT_SIGNATURE,
                                            ]
                                        )
                                    ),
                                    func.concat(
                                        literal("dataflow_from_value"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                # Edges from dataflow signature nodes to memory
                                (
                                    (
                                        current_edge.kind.in_(
                                            [
                                                EdgeKind.DATAFLOW_SIGNATURE,
                                                EdgeKind.DIRECT_DATAFLOW_SIGNATURE,
                                            ]
                                            + (
                                                []
                                                if thin
                                                else [EdgeKind.INDIRECT_DATAFLOW_SIGNATURE]
                                            )
                                        )
                                        & (target_node.kind == NodeKind.MEMORY_LOCATION)
                                    ),
                                    func.concat(
                                        literal("dataflow_to_memory"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                # Edges from dataflow signature nodes to values
                                (
                                    (
                                        current_edge.kind.in_(
                                            [
                                                EdgeKind.DATAFLOW_SIGNATURE,
                                                EdgeKind.DIRECT_DATAFLOW_SIGNATURE,
                                            ]
                                            + (
                                                []
                                                if thin
                                                else [EdgeKind.INDIRECT_DATAFLOW_SIGNATURE]
                                            )
                                        )
                                    ),
                                    func.concat(
                                        literal("dataflow_to_value"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                (
                                    current_edge.kind == EdgeKind.LOAD_MEMORY,
                                    func.concat(
                                        literal("load"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                (
                                    current_edge.kind == EdgeKind.STORE_MEMORY,
                                    func.concat(
                                        literal("store"),
                                        ":",
                                        current_edge.attributes["context"].astext,
                                    ),
                                ),
                                (
                                    current_edge.kind == EdgeKind.OPERAND_TO_PARAM_BINDING,
                                    literal("operand_to_param_binding"),
                                ),
                                (
                                    current_edge.kind == EdgeKind.PARAM_BINDING_TO_ARG,
                                    func.concat(
                                        literal("call"),
                                        ":",
                                        current_call.uuid,
                                        ":",
                                        current_edge.attributes["caller_context"].astext,
                                        "-->",
                                        current_edge.attributes["callee_context"].astext,
                                    ),
                                ),
                                (
                                    current_edge.kind == EdgeKind.RETURN_VALUE_TO_CALL_RETURN,
                                    literal("return_value_to_call_return"),
                                ),
                                (
                                    current_edge.kind == EdgeKind.CALL_RETURN_TO_CALLER,
                                    func.concat(
                                        literal("return"),
                                        ":",
                                        current_call.uuid,
                                        ":",
                                        current_edge.attributes["caller_context"].astext,
                                        "-->",
                                        current_edge.attributes["callee_context"].astext,
                                    ),
                                ),
                            ],
                        ).label("symbol"),
                    ]
                )
                .select_from(
                    join(
                        join(
                            join(
                                join(
                                    current_edge,
                                    target_node,
                                    (current_edge.target == target_node.uuid),
                                ),
                                source_node,
                                (current_edge.source == source_node.uuid),
                            ),
                            current_call_edge,
                            (
                                (current_edge.kind == EdgeKind.PARAM_BINDING_TO_ARG)
                                & (current_edge.source == current_call_edge.target)
                                & (current_call_edge.kind == EdgeKind.CALL_TO_PARAM_BINDING)
                            ),
                            isouter=True,
                        ),
                        current_call,
                        (
                            (
                                (current_call_edge.uuid != None)
                                & (current_call_edge.source == current_call.uuid)
                            )
                            | (
                                (current_call_edge.uuid == None)
                                & (current_edge.target == current_call.uuid)
                            )
                        ),
                    )
                )
                .alias()
            )

            graph.session.execute(
                insert(cls.edge_symbol_table).from_select(
                    ["uuid", "source", "target", "symbol"],
                    select(
                        [
                            edge_symbol_query.c.uuid,
                            edge_symbol_query.c.source,
                            edge_symbol_query.c.target,
                            edge_symbol_query.c.symbol,
                        ]
                    ).where(edge_symbol_query.c.symbol != None),
                )
            )
            cls.update_edge_symbol_table_statistics(graph)
            graph.session.commit()

        @classmethod
        def populate_transition_table(cls, graph: CPG) -> None:
            if not cls._create_transition_table(graph):
                return

            callsite_stack_symbols = cls.select_callsite_stack_symbols(graph)
            callsite_stack_symbols2 = aliased(callsite_stack_symbols)
            context_stack_symbols = cls.select_context_stack_symbols(graph)

            all_stack_symbols = union(
                select([callsite_stack_symbols.c.stack_symbol]).select_from(callsite_stack_symbols),
                select([context_stack_symbols.c.stack_symbol]).select_from(context_stack_symbols),
                select([literal(BOT).label("stack_symbol")]),
            ).alias()

            # NOTE(lb): Any update to this query should be reflected in the class
            # docstring.
            rules = {
                "intraprocedural": select(
                    [
                        literal("DATAFLOW"),
                        all_stack_symbols.c.stack_symbol,
                        literal("value_definition_to_use"),
                        literal("DATAFLOW"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
                "operand_to_param_binding": select(
                    [
                        literal("DATAFLOW"),
                        all_stack_symbols.c.stack_symbol,
                        literal("operand_to_param_binding"),
                        literal("DATAFLOW"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
                "return_value_to_call_return": select(
                    [
                        literal("DATAFLOW"),
                        all_stack_symbols.c.stack_symbol,
                        literal("return_value_to_call_return"),
                        literal("DATAFLOW"),
                        array([all_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                    ],
                ).select_from(all_stack_symbols),
                "param_binding_to_arg": (
                    select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols.c.stack_symbol,
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([]).cast(ARRAY(String)),
                        ],
                    ).select_from(callsite_stack_symbols)
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols2.c.stack_symbol,
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array(
                                [
                                    callsite_stack_symbols.c.stack_symbol,
                                    callsite_stack_symbols2.c.stack_symbol,
                                ]
                            ).cast(ARRAY(String)),
                        ],
                    ).select_from(
                        join(
                            callsite_stack_symbols,
                            callsite_stack_symbols2,
                            (callsite_stack_symbols2.c.callee == callsite_stack_symbols.c.caller)
                            & (
                                callsite_stack_symbols2.c.callee_context
                                == callsite_stack_symbols.c.caller_context
                            ),
                        )
                    )
                ),
                "empty_param_binding_to_arg": (
                    select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.caller_context, BOT]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(callsite_stack_symbols)
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.stack_symbol, BOT]).cast(ARRAY(String)),
                        ],
                    ).select_from(callsite_stack_symbols)
                ),
                "context_param_binding_to_arg": (
                    select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols.c.callee_context,
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.caller_context]).cast(ARRAY(String)),
                        ],
                    ).select_from(callsite_stack_symbols)
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            context_stack_symbols.c.stack_symbol,
                            func.concat(
                                literal("call"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array(
                                [
                                    callsite_stack_symbols.c.stack_symbol,
                                    context_stack_symbols.c.stack_symbol,
                                ]
                            ).cast(ARRAY(String)),
                        ],
                    ).select_from(
                        join(
                            context_stack_symbols,
                            callsite_stack_symbols,
                            (
                                context_stack_symbols.c.stack_symbol
                                == callsite_stack_symbols.c.caller_context
                            )
                            & (context_stack_symbols.c.function == callsite_stack_symbols.c.caller),
                        )
                    )
                ),
                "call_return_to_caller": (
                    select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols2.c.stack_symbol,
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array(
                                [
                                    callsite_stack_symbols.c.stack_symbol,
                                    callsite_stack_symbols2.c.stack_symbol,
                                ]
                            ).cast(ARRAY(String)),
                        ],
                    ).select_from(
                        join(
                            callsite_stack_symbols,
                            callsite_stack_symbols2,
                            (callsite_stack_symbols2.c.callee == callsite_stack_symbols.c.caller)
                            & (
                                callsite_stack_symbols2.c.callee_context
                                == callsite_stack_symbols.c.caller_context
                            ),
                        )
                    )
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols.c.stack_symbol,
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([]).cast(ARRAY(String)),
                        ],
                    ).select_from(callsite_stack_symbols)
                ),
                "empty_call_return_to_caller": (
                    select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(callsite_stack_symbols)
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.caller_context, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(callsite_stack_symbols)
                ),
                "context_call_return_to_caller": (
                    select(
                        [
                            literal("DATAFLOW"),
                            context_stack_symbols.c.stack_symbol,
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array(
                                [
                                    callsite_stack_symbols.c.callee_context,
                                    context_stack_symbols.c.stack_symbol,
                                ]
                            ).cast(ARRAY(String)),
                        ],
                    ).select_from(
                        join(
                            context_stack_symbols,
                            callsite_stack_symbols,
                            (
                                context_stack_symbols.c.stack_symbol
                                == callsite_stack_symbols.c.caller_context
                            )
                            & (context_stack_symbols.c.function == callsite_stack_symbols.c.caller),
                        )
                    )
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            callsite_stack_symbols.c.callee_context,
                            func.concat(
                                literal("return"),
                                ":",
                                callsite_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([callsite_stack_symbols.c.caller_context]).cast(ARRAY(String)),
                        ],
                    ).select_from(callsite_stack_symbols)
                ),
                "load_memory": (
                    union(
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("load"), ":", callsite_stack_symbols.c.callee_context
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("load"), ":", context_stack_symbols.c.stack_symbol
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("load"), ":", context_stack_symbols.c.stack_symbol
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                    )
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(literal("load"), ":", context_stack_symbols.c.stack_symbol),
                            literal("DATAFLOW"),
                            array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(context_stack_symbols)
                ),
                "store_memory": (
                    select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("store"), ":", context_stack_symbols.c.stack_symbol
                            ),
                            literal("DATAFLOW"),
                            array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(context_stack_symbols)
                    if cls.is_reversed
                    else union(
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("store"), ":", callsite_stack_symbols.c.callee_context
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("store"), ":", context_stack_symbols.c.stack_symbol
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("store"), ":", context_stack_symbols.c.stack_symbol
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                    )
                ),
                "dataflow_from_memory": (
                    union(
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_memory"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_from_memory"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_memory"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                    )
                    if cls.is_reversed
                    else select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("dataflow_from_memory"),
                                ":",
                                context_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(context_stack_symbols)
                ),
                "dataflow_to_memory": (
                    select(
                        [
                            literal("DATAFLOW"),
                            literal(BOT),
                            func.concat(
                                literal("dataflow_to_memory"),
                                ":",
                                context_stack_symbols.c.stack_symbol,
                            ),
                            literal("DATAFLOW"),
                            array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                ARRAY(String)
                            ),
                        ],
                    ).select_from(context_stack_symbols)
                    if cls.is_reversed
                    else union(
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_memory"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_to_memory"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_memory"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([literal(BOT)]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                    )
                ),
                "dataflow_from_value": (
                    union(
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                    ARRAY(String)
                                ),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([callsite_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                    )
                    if cls.is_reversed
                    else union(
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([callsite_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol, BOT]).cast(
                                    ARRAY(String)
                                ),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_from_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                    )
                ),
                "dataflow_to_value": (
                    union(
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([callsite_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol, BOT]).cast(
                                    ARRAY(String)
                                ),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                    )
                    if cls.is_reversed
                    else union(
                        select(
                            [
                                literal("DATAFLOW"),
                                context_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                literal(BOT),
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    context_stack_symbols.c.stack_symbol,
                                ),
                                literal("DATAFLOW"),
                                array([context_stack_symbols.c.stack_symbol, literal(BOT)]).cast(
                                    ARRAY(String)
                                ),
                            ],
                        ).select_from(context_stack_symbols),
                        select(
                            [
                                literal("DATAFLOW"),
                                callsite_stack_symbols.c.stack_symbol,
                                func.concat(
                                    literal("dataflow_to_value"),
                                    ":",
                                    callsite_stack_symbols.c.callee_context,
                                ),
                                literal("DATAFLOW"),
                                array([callsite_stack_symbols.c.stack_symbol]).cast(ARRAY(String)),
                            ],
                        ).select_from(callsite_stack_symbols),
                    )
                ),
            }

            graph.session.execute(
                insert(cls.transition_table).from_select(
                    ["old_state", "old_stack", "input", "new_state", "new_stack"],
                    union(*[rules[i] for i in rules]),
                )
            )
            cls.update_transition_table_statistics(graph)
            graph.session.commit()

    return CSDataflowPath


CSThinDataflowPath: Type[Path] = ConfigureDFGPath(thin=True)

CSDataflowPath: Type[Path] = ConfigureDFGPath(thin=False)


class CallGraphPath(CFLPath):
    @classmethod
    def table_name_prefix(cls) -> str:
        # The PDA tables are the same for all instances
        return "callgraph"

    @classmethod
    def initial_state(cls) -> ClauseElement:
        return literal("EMPTY").cast(String)

    @classmethod
    def initial_stack(cls) -> ClauseElement:
        return array([BOT]).cast(ARRAY(String))

    @classmethod
    def populate_edge_symbol_table(cls, graph: CPG) -> None:
        """Preprocess the graph's edge table and give each edge a symbol to be used during during
        CFL-based path extension."""

        if not cls._create_edge_symbol_table(graph):
            return

        # NOTE(lb): Any update to this query should be reflected in the class
        # docstring.
        edge_symbol_query = (
            select(
                [
                    graph.Edge.uuid,
                    graph.Edge.source,
                    graph.Edge.target,
                    func.concat(
                        graph.Edge.attributes["caller_context"].astext,
                        literal("-->"),
                        graph.Edge.attributes["callee_context"].astext,
                    ).label("symbol"),
                ]
            )
            .where(graph.Edge.kind == EdgeKind.CALLGRAPH)
            .alias()
        )

        graph.session.execute(
            insert(cls.edge_symbol_table).from_select(
                ["uuid", "source", "target", "symbol"],
                select(
                    [
                        edge_symbol_query.c.uuid,
                        edge_symbol_query.c.source,
                        edge_symbol_query.c.target,
                        edge_symbol_query.c.symbol,
                    ]
                ),
            )
        )
        cls.update_edge_symbol_table_statistics(graph)
        graph.session.commit()

    @classmethod
    def populate_transition_table(cls, graph: CPG) -> None:
        if not cls._create_transition_table(graph):
            return

        callgraph_edge_symbols = (
            select(
                [
                    graph.Edge.attributes["caller_context"].astext.label("caller"),
                    graph.Edge.attributes["callee_context"].astext.label("callee"),
                    func.concat(
                        graph.Edge.attributes["caller_context"].astext,
                        literal("-->"),
                        graph.Edge.attributes["callee_context"].astext,
                    ).label("symbol"),
                ]
            )
            .where(graph.Edge.kind == EdgeKind.CALLGRAPH)
            .alias()
        )

        # NOTE(lb): Any update to this query should be reflected in the class
        # docstring.
        rules = {
            "empty": select(
                [
                    literal("EMPTY"),
                    literal(BOT),
                    callgraph_edge_symbols.c.symbol,
                    literal("CONTEXT"),
                    (
                        array([callgraph_edge_symbols.c.caller]).cast(ARRAY(String))
                        if cls.is_reversed
                        else array([callgraph_edge_symbols.c.callee]).cast(ARRAY(String))
                    ),
                ]
            ).select_from(callgraph_edge_symbols),
            "context": select(
                [
                    literal("CONTEXT"),
                    (
                        callgraph_edge_symbols.c.callee
                        if cls.is_reversed
                        else callgraph_edge_symbols.c.caller
                    ),
                    callgraph_edge_symbols.c.symbol,
                    literal("CONTEXT"),
                    (
                        array([callgraph_edge_symbols.c.caller]).cast(ARRAY(String))
                        if cls.is_reversed
                        else array([callgraph_edge_symbols.c.callee]).cast(ARRAY(String))
                    ),
                ]
            ).select_from(callgraph_edge_symbols),
        }

        graph.session.execute(
            insert(cls.transition_table).from_select(
                ["old_state", "old_stack", "input", "new_state", "new_stack"],
                union(*[rules[i] for i in rules]),
            )
        )
        cls.update_transition_table_statistics(graph)
        graph.session.commit()



"""Nodes corresponding to translation units in the compilation process.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""

from mate_common.models.cpg_types import NodeKind
from mate_query.cpg.models.node._typechecking import NodeMixin


class TranslationUnit(NodeMixin):
    _kind = NodeKind.TRANSLATION_UNIT



"""Core node models and functions.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import Column, Index, text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.types import Enum as SAEnum

from mate_common.assertions import mate_assert
from mate_common.models.cpg_types.mate import NODE_PROVENANCE, EdgeKind, MATEComponent, NodeKind
from mate_query.cpg.models.core.base import Base
from mate_query.cpg.models.core.relationships import (
    add_properties_to_node_class,
    make_edge_relationship,
)

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple, Type

    from sqlalchemy.orm import RelationshipProperty

    from mate_query.cpg.models.core.relationships import Direction


def node_kind_column() -> Column:
    return Column(
        "kind",
        SAEnum(NodeKind, values_callable=lambda x: [e.value for e in x], validate_strings=True),
        index=True,
        nullable=False,
    )


class BaseNode(Base):
    # NOTE(lb): This is shared between SQLite (declarative) and Postgres
    # (imperative) mappings.
    # TODO(#1044): Remove me!

    kind = node_kind_column()

    @classmethod
    def from_filename_line_no(cls, filepath: str, line_no: Optional[int] = None) -> Node:
        """Returns Nodes whose debug symbols contain the given filepath, or the matching filepath
        and line number."""
        session = cls._session().query(cls)
        if line_no:
            return session.filter(
                Node.attributes["location"]["file"].as_string() == filepath,
                Node.attributes["location"]["line"].as_integer() == line_no,
            ).all()
        return session.filter(Node.attributes["location"]["file"].as_string() == filepath).all()

    @classmethod
    def from_dirpath(cls, dirpath: str) -> List[Node]:
        """Returns Nodes whose debug symbols contain a filepath that includes the given dirpath."""
        return (
            cls._session()
            .query(cls)
            .filter(Node.attributes["location"]["file"].as_string().like("%" + dirpath + "/%"))
            .all()
        )

    @hybrid_property
    def provenance(self) -> MATEComponent:
        return NODE_PROVENANCE[self.kind]

    @property
    def location_string(self) -> str:
        """Return a human-readable source location string for this node."""
        # TODO: This should be optimized and specialized more.
        location = self.attributes.get("location")
        if location is None:
            return "<unknown>"
        filename = location.get("file", "<unknown>")
        line = location.get("line", "<unknown>")
        column = location.get("column")
        try:
            function = f":{self.parent_block.parent_function.demangled_name}"
        except:
            function = ""
        if column is not None:
            return f"{filename}:{line}:{column}{function}"
        else:
            return f"{filename}:{line}{function}"

    def __repr__(self) -> str:
        try:
            return f"<{self.__class__._kind.value}({self.uuid})>"
        except AttributeError:
            return f"<Node({self.uuid})>"

    @classmethod
    def edge_relationship(
        cls,
        edge_kind: EdgeKind,
        other: Type[Node],
        direction: Optional[Direction] = None,
        **kwargs: Any,
    ) -> RelationshipProperty:
        return make_edge_relationship(
            cls.cpg, edge_kind, cls, other, direction, **kwargs  # type: ignore[arg-type]
        )

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """This function sets up properties like ``Instruction.opcode`` and relationships like
        ``Instruction.parent_block`` that can be used both inside SQL expression contexts (queries),
        and outside of them in Python-land.

        In detail:

        * Add a docstring to subclasses which is taken from the "description"
        field of the JSON schema for this node's kind (in ``nodes.json``).
        * Add SQLAlchemy relationships for each item in the subclass's
        ``_edges`` attribute. These relationships are automatically
        configured according to the ``endpoints.json`` and ``relationships``
        schemata.
        * Add a ``hybrid_property`` for each property in the JSON schema for
        this node kind.

            + In Python expression contexts, the ``hybrid_property`` will return
            an enum value if relevant, or just the JSON value
            + In SQL expression contexts, the ``hybrid_property`` will wrap the
            value in one of SQLAlchemy's scalar cast methods (e.g. ``as_string()``)
        """
        super().__init_subclass__(**kwargs)
        if getattr(cls, "_kind", None) is not None:
            add_properties_to_node_class(cls)  # type: ignore[arg-type]

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        try:
            return {"polymorphic_identity": cls._kind}
        except AttributeError:
            return {"polymorphic_on": cls.kind, "polymorphic_identity": None}

    @declared_attr
    def __table_args__(cls) -> Tuple[Index, ...]:
        args = [
            Index(
                f"node_{cls.build_id}_location",
                text("(((attributes -> 'location') ->> 'file'))"),
                text("((((attributes -> 'location') ->> 'line')::int))"),
            ),
            Index(
                f"node_{cls.build_id}_allocation",
                text("((attributes ->> 'alias_set_identifier'))"),
            ),
        ]
        return tuple(args)


class Node(BaseNode):
    def __init__(
        self,
        uuid: str,  # pylint: disable=unused-argument
        kind: NodeKind,  # pylint: disable=unused-argument
        attributes: Dict[str, Any] = dict(),  # pylint: disable=unused-argument
    ):
        mate_assert(False, "Unused. Just exists to make Mypy happy.")



"""Nodes in the LLVM AST.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""
from __future__ import annotations

import binascii
import logging
import typing
from collections.abc import Iterable
from typing import TYPE_CHECKING, Type

from sqlalchemy import and_
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import Comparator, hybrid_method, hybrid_property
from sqlalchemy.orm import backref, relationship
from sqlalchemy.types import JSON

from mate_common.datastructures.digraph import Digraph
from mate_common.models.analyses import SalientFunction
from mate_common.models.cpg_types import EdgeKind, NodeKind
from mate_query.cpg.models.core.relationships import Direction
from mate_query.cpg.models.node._typechecking import NodeMixin

if TYPE_CHECKING:
    from typing import Any, ClassVar, Dict, List, Optional

    from sqlalchemy.orm import RelationshipProperty

    from mate_query.cpg.models.core.cpg import MemoryLocation
    from mate_query.cpg.models.core.node import Node
    from mate_query.db import Graph as CPG


logger = logging.getLogger(__name__)


class LLVMType(NodeMixin):
    _kind = NodeKind.LLVM_TYPE

    @declared_attr
    def instructions(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.HAS_LLVM_TYPE, cls.cpg.Instruction, backref="llvm_type"
        )

    @declared_attr
    def functions(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.HAS_LLVM_TYPE, cls.cpg.Function, backref="llvm_type")

    @declared_attr
    def blocks(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.HAS_LLVM_TYPE, cls.cpg.Block, backref="llvm_type")

    @declared_attr
    def constants(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.HAS_LLVM_TYPE, cls.cpg.Constant, backref="llvm_type")

    @declared_attr
    def arguments(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.HAS_LLVM_TYPE, cls.cpg.Argument, backref="llvm_type")

    @declared_attr
    def global_variables(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.HAS_LLVM_TYPE, cls.cpg.GlobalVariable, backref="llvm_type"
        )

    @hybrid_property
    def name(self):
        return self.attributes["definition"]["name"]

    @name.expression  # type: ignore[no-redef]
    def name(cls):
        return cls.attributes["definition"]["name"].as_string()

    @hybrid_property
    def definition(self) -> Any:
        return self.attributes["definition"]

    def get(self, key: str) -> Any:
        try:
            return self.definition.get(key)
        except AttributeError:  # definition is a string, e.g. "void"
            return None

    @hybrid_property
    def is_void(self) -> bool:
        return self.definition.as_string() == "void"

    @hybrid_property
    def is_array_type(self) -> bool:
        return self.get("array") is not None

    @is_array_type.expression  # type: ignore[no-redef]
    def is_array_type(cls):
        return cls.definition["array_size"].as_integer() != JSON.NULL

    @hybrid_property
    def array_size(self) -> int:
        return self.attributes["definition"]["array_size"]

    @hybrid_property
    def is_pointer_type(self) -> bool:
        return self.get("pointer") is not None

    @is_pointer_type.expression  # type: ignore[no-redef]
    def is_pointer_type(cls):
        return cls.definition.has_key("pointer")

    @hybrid_property
    def is_function_type(self) -> bool:
        return self.is_pointer_type and self.get("pointer").get("function") is not None

    @is_function_type.expression  # type: ignore[no-redef]
    def is_function_type(cls):
        return cls.definition["pointer"].has_key("function")

    @hybrid_property
    def is_struct_type(self) -> bool:
        return self.get("struct") is not None

    @is_struct_type.expression  # type: ignore[no-redef]
    def is_struct_type(cls):
        return cls.definition.has_key("struct")

    @hybrid_property
    def is_named_struct_type(self):
        return self.is_struct_type and self.get("name") is not None

    @is_named_struct_type.expression  # type: ignore[no-redef]
    def is_named_struct_type(cls):
        return cls.is_struct_type & cls.definition.has_key("name")

    @hybrid_property
    def size_in_bits(self):
        return self.attributes["size_in_bits"]

    @size_in_bits.expression  # type: ignore[no-redef]
    def size_in_bits(cls):
        return cls.attributes["size_in_bits"].as_integer()


class Constant(NodeMixin):
    _kind = NodeKind.CONSTANT


class ConstantInt(Constant):
    _kind = NodeKind.CONSTANT_INT


class ConstantFP(Constant):
    _kind = NodeKind.CONSTANT_FP


class CSComparator(Comparator):
    """A specialized SQLAlchemy Comparator for ``ConstantString``"""

    @staticmethod
    def __cast__(value: Any) -> Any:
        """Turns value into a hexstring, if appropriate."""
        if isinstance(value, str):
            return binascii.hexlify(value.encode()).decode()
        elif isinstance(value, bytes):
            return binascii.hexlify(value).decode()
        elif isinstance(value, Iterable):
            return [CSComparator.__cast__(v) for v in value]
        return value

    def operate(self, op: Any, other: Any, **_kwargs: Any) -> Any:
        return op(self.__clause_element__(), CSComparator.__cast__(other))


class ConstantString(Constant):
    _kind = NodeKind.CONSTANT_STRING

    @hybrid_property
    def string_value(self):
        return binascii.unhexlify(self.attributes["string_value"])

    @string_value.comparator  # type: ignore[no-redef]
    def string_value(cls):
        return CSComparator(cls.attributes["string_value"].as_string())


class ConstantUndef(Constant):
    _kind = NodeKind.CONSTANT_UNDEF


class Variable(NodeMixin):
    _kind = NodeKind.VARIABLE


class Argument(Variable):
    _kind = NodeKind.ARGUMENT

    @declared_attr
    def as_dwarf(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.ARGUMENT_TO_DWARF_ARGUMENT, cls.cpg.DWARFArgument)

    @property
    def memory_locations(cls) -> Optional[List[MemoryLocation]]:
        if not cls.allocation_site:
            return None
        return cls.allocation_site.allocates

    @declared_attr
    def allocation_site(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Alloca,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.target == cls.uuid,
                cls.cpg.Edge.kind == EdgeKind.CREATES_VAR,
            ),
            secondaryjoin=cls.cpg.Alloca.uuid == cls.cpg.Edge.source,
            uselist=False,
            doc="Link to the Alloca that created this Argument",
        )


class Function(NodeMixin):
    _kind = NodeKind.FUNCTION

    @declared_attr
    def arguments(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.FUNCTION_TO_ARGUMENT, cls.cpg.Argument, backref="parent_function"
        )

    @declared_attr
    def blocks(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.BLOCK_TO_PARENT_FUNCTION, cls.cpg.Block, backref="parent_function"
        )

    @declared_attr
    def machine_functions(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.MI_FUNCTION_TO_IR_FUNCTION,
            cls.cpg.MachineFunction,
            backref="ir_function",
        )

    @declared_attr
    def plt_stub(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.FUNCTION_TO_PLT_STUB,
            cls.cpg.PLTStub,
            backref="ir_function",
        )

    @declared_attr
    def entry_block(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.FUNCTION_TO_ENTRY_BLOCK, cls.cpg.Block)

    @declared_attr
    def entry_controls_blocks(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.FUNCTION_ENTRY_TO_CONTROL_DEPENDENT_BLOCK,
            cls.cpg.Block,
            direction=Direction.OUT,
            backref="controlled_by_function_entry",
        )

    @declared_attr
    def entry_controls_instructions(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.FUNCTION_ENTRY_TO_CONTROL_DEPENDENT_INSTRUCTION,
            cls.cpg.Instruction,
            direction=Direction.OUT,
            backref="controlled_by_function_entry",
        )

    @declared_attr
    def callsites(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.CALL_TO_FUNCTION, cls.cpg.CallSite, backref="callees")

    @declared_attr
    def callees(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.CALLGRAPH,
            cls,  # type: ignore[arg-type]
            direction=Direction.OUT,
            backref="callers",
        )

    @declared_attr
    def signatures(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=cls.uuid == cls.cpg.Edge.target,
            secondaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DATAFLOW_SIGNATURE_FOR_FUNCTION,
                cls.cpg.Edge.source == cls.cpg.Node.uuid,
            ),
            lazy="dynamic",
            doc="Connects a function to signatures defined for its call sites.",
        )

    def as_salient(self) -> SalientFunction:
        """Create a ``SalientFunction`` from this ``Function``."""
        return SalientFunction(cpg_id=self.uuid, demangled_name=self.demangled_name, name=self.name)


class GlobalVariable(Variable):
    _kind = NodeKind.GLOBAL_VARIABLE

    @declared_attr
    def initializer(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.source == cls.uuid,
                cls.cpg.Edge.kind == EdgeKind.GLOBAL_TO_INITIALIZER,
            ),
            secondaryjoin=cls.cpg.Edge.target == cls.cpg.Node.uuid,
            lazy="select",
            uselist=False,
            backref=backref(
                "initializes", doc="Connects an initializer to the global variables it initializes."
            ),
            doc="Connects a global variable to its initializer.",
        )

    @property
    def memory_locations(cls) -> Optional[List[MemoryLocation]]:
        return cls.allocates


class Module(NodeMixin):
    _kind = NodeKind.MODULE

    def translation_units(cls) -> RelationshipProperty:
        """A sequence of all source-level translation units in this module."""
        return cls.edge_relationship(EdgeKind.MODULE_TO_TRANSLATION_UNIT, cls.cpg.TranslationUnit)


class Block(NodeMixin):

    _kind = NodeKind.BLOCK

    @declared_attr
    def mi_blocks(cls) -> RelationshipProperty:
        """All middle-end blocks (``MachineBasicBlock``s) for this IR block."""
        return cls.edge_relationship(
            EdgeKind.MI_BLOCK_TO_IR_BLOCK, cls.cpg.MachineBasicBlock, backref="ir_block"
        )

    @declared_attr
    def instructions(cls) -> RelationshipProperty:
        """All instructions in this block."""
        return cls.edge_relationship(
            EdgeKind.INSTRUCTION_TO_PARENT_BLOCK,
            cls.cpg.Instruction,
            backref="parent_block",
        )

    @declared_attr
    def entry(cls) -> RelationshipProperty:
        """The entry instruction for this block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION, cls.cpg.Instruction)

    @declared_attr
    def terminator(cls) -> RelationshipProperty:
        """The terminator instruction for this block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_TERMINATOR_INSTRUCTION, cls.cpg.Instruction)

    @declared_attr
    def successors(cls) -> RelationshipProperty:
        """The successor blocks for this block."""
        return cls.edge_relationship(
            EdgeKind.BLOCK_TO_SUCCESSOR_BLOCK,
            cls,  # type: ignore[arg-type]
            direction=Direction.OUT,
            backref="predecessors",
        )

    @declared_attr
    def controls(cls) -> RelationshipProperty:
        """The blocks controlled by this block."""
        return cls.edge_relationship(
            EdgeKind.BLOCK_TO_CONTROL_DEPENDENT_BLOCK,
            cls,  # type: ignore[arg-type]
            direction=Direction.OUT,
            backref="controlled_by",
        )

    def sorted_mi_blocks(self) -> typing.Iterable[Node]:
        """The ``MachineBasicBlock``s for this block, sorted."""
        # NOTE(lb): We don't use standard Python `sorted` here because the
        # (assumed) complete linear order by control-flow is very expensive to
        # compute (it requires SQL path queries). Instead, we assume each MI
        # block will be an immediate successor of another one, or the last
        # block in the list, and perform a topological sort based on the
        # (partial) successor relation.
        #
        # The point is to optimize for number of SQL queries.
        #
        # This is still pretty expensive.
        if len(self.mi_blocks) <= 1:
            return self.mi_blocks

        def get_successors(block: Node) -> typing.Iterable[Node]:
            return (
                e.target_node
                for e in block.outgoing
                if e.kind == EdgeKind.BLOCK_TO_SUCCESSOR_BLOCK and e in self.mi_blocks
            )

        done = Digraph.from_successor_function(self.mi_blocks, get_successors).topological_sort()

        if len(done) != len(self.mi_blocks):
            logger.error(f"Weird list of MI blocks for {self}. Sorted: {done}")

        return done


def _make_use_relationship(
    cpg: CPG,
    this_class: Type[Node],
    constraints: List[Any],
    other: Optional[Type[Node]] = None,
    doc: Optional[str] = None,
) -> RelationshipProperty:
    if other is None:
        other = cpg.Node
    return relationship(
        other,
        secondary=cpg.Edge.__table__,
        primaryjoin=and_(
            *(
                [
                    cpg.Edge.target == this_class.uuid,
                    cpg.Edge.kind == EdgeKind.VALUE_DEFINITION_TO_USE.value,
                ]
                + constraints
            )
        ),
        secondaryjoin=other.uuid == cpg.Edge.source,
        lazy="select",
        uselist=False,
        doc=doc,
    )


def _make_operand_relationship(
    cpg: CPG,
    this_class: Type[Node],
    number: int,
    other: Optional[Type[Node]] = None,
    additional_constraints: List[Any] = [],
    doc: Optional[str] = None,
) -> RelationshipProperty:
    return _make_use_relationship(
        cpg,
        this_class,
        [cpg.Edge.attributes["operand_number"].as_integer() == number] + additional_constraints,
        other,
        doc=doc if doc is not None else f"Operand {number} of the instruction.",
    )


class Instruction(NodeMixin):
    _kind = NodeKind.INSTRUCTION

    @declared_attr
    def controls(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.TERMINATOR_INSTRUCTION_TO_CONTROL_DEPENDENT_INSTRUCTION,
            cls,  # type: ignore[arg-type]
            direction=Direction.OUT,
            backref="controlled_by",
        )

    @declared_attr
    def successors(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.INSTRUCTION_TO_SUCCESSOR_INSTRUCTION,
            cls,  # type: ignore[arg-type]
            direction=Direction.OUT,
            backref="predecessors",
        )

    @declared_attr
    def uses(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.target == cls.uuid,
            ),
            secondaryjoin=and_(
                cls.cpg.Edge.source == cls.uuid,
                cls.cpg.Edge.kind == EdgeKind.VALUE_DEFINITION_TO_USE,
            ),
            lazy="dynamic",
            backref=backref(
                "used_by", doc="Connects a definition to the instructions that use it."
            ),
            doc="Connects an instruction to the definitions it uses.",
        )

    @declared_attr
    def operand0(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)

    @declared_attr
    def operand1(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 1)

    @declared_attr
    def operand2(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 2)

    @declared_attr
    def operand3(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 3)

    @declared_attr
    def operand4(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 4)

    @declared_attr
    def operand5(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 5)

    @declared_attr
    def operand6(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 6)

    @declared_attr
    def operand7(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 7)

    @declared_attr
    def operand8(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 8)

    @declared_attr
    def operand9(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 9)

    @classmethod
    def operand_relationship(
        cls, cpg: CPG, operand_number: int, operand_class: Optional[Type[Node]] = None
    ) -> RelationshipProperty:
        """Specifications of the operands of an LLVM instruction.

        Arguments:

        * ``cpg``: The CPG

        * ``operand_number``: The operand number

        * ``operand_class``: The ``Node`` subclass that should be on the other end
          of the use edge/relationship, if it's more specific than LLVM's ``Value``
          class (e.g. the first operand of a ``CallSite``) node should be a
          ``Function``, but the other operands could be any ``Node`` in the LLVM
          AST).
        """
        return _make_operand_relationship(cpg, cls, operand_number, operand_class)


class Alloca(Instruction):
    _kind = NodeKind.ALLOCA

    @declared_attr
    def variable(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Variable,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.source == cls.uuid,
                cls.cpg.Edge.kind == EdgeKind.CREATES_VAR,
            ),
            secondaryjoin=cls.cpg.Edge.target == cls.cpg.Node.uuid,
            doc="Link to the Variable created by this alloca (if one was created)",
            uselist=False,
        )


class Load(Instruction):
    _kind = NodeKind.LOAD

    @declared_attr
    def pointer_operand(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)


class Store(Instruction):
    _kind = NodeKind.STORE

    @declared_attr
    def value_operand(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)

    @declared_attr
    def pointer_operand(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 1)


class Ret(Instruction):
    _kind = NodeKind.RET

    @declared_attr
    def value(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)


class Resume(Instruction):
    _kind = NodeKind.RESUME

    @declared_attr
    def value(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)


class CallSite(Instruction):
    @hybrid_method
    def calls(self, *fns):
        return any(callee in self.callees for callee in fns)

    @calls.expression  # type: ignore[no-redef]
    def calls(cls, *fns):
        return cls.callees.any(cls.cpg.Function.name.in_(fns))

    @declared_attr
    def signatures(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Node,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=cls.uuid == cls.cpg.Edge.target,
            secondaryjoin=and_(
                cls.cpg.Edge.kind == EdgeKind.DATAFLOW_SIGNATURE_FOR_CALLSITE,
                cls.cpg.Edge.source == cls.cpg.Node.uuid,
            ),
            lazy="dynamic",
            doc="Connects a callsite to signatures defined for its callees.",
        )

    @declared_attr
    def argument0(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 0)

    @declared_attr
    def argument1(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 1)

    @declared_attr
    def argument2(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 2)

    @declared_attr
    def argument3(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 3)

    @declared_attr
    def argument4(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 4)

    @declared_attr
    def argument5(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 5)

    @declared_attr
    def argument6(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 6)

    @declared_attr
    def argument7(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 7)

    @declared_attr
    def argument8(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 8)

    @declared_attr
    def argument9(cls) -> RelationshipProperty:
        return cls.argument_relationship(cls.cpg, 9)

    @declared_attr
    def callee_operand(cls) -> RelationshipProperty:
        return _make_use_relationship(
            cls.cpg,
            cls,  # type: ignore[arg-type]
            [cls.cpg.Edge.attributes["is_callee"].as_boolean() == True],
            doc="The callsite's callee operand.",
        )

    @classmethod
    def argument_relationship(cls, cpg: CPG, number: int) -> RelationshipProperty:
        return _make_operand_relationship(
            cpg,
            cls,
            number,
            additional_constraints=[
                cpg.Edge.attributes["is_argument_operand"].as_boolean() == True
            ],
            doc=f"Argument {number} of the callsite.",
        )

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        # NOTE(lb): This class should never be mapped to directly, only as a
        # union of its subclasses.
        if cls._kind == NodeKind.INSTRUCTION:
            return {"polymorphic_identity": None}
        return {"polymorphic_identity": cls._kind}


class Call(CallSite):
    _kind = NodeKind.CALL


class Invoke(CallSite):
    _kind = NodeKind.INVOKE


class Intrinsic(Call):

    overloaded: ClassVar[bool] = False

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        # NOTE(lb): This class should never be mapped to directly, only as a
        # union of its subclasses.
        if cls._kind == NodeKind.CALL:
            return {"polymorphic_identity": None}
        return {"polymorphic_identity": cls._kind}


class Memcpy(Intrinsic):

    _kind = NodeKind.MEMCPY

    overloaded = True

    @declared_attr
    def dest(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)

    @declared_attr
    def src(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 1)

    @declared_attr
    def size(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 2)

    @declared_attr
    def volatile(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 3)


class Memset(Intrinsic):

    _kind = NodeKind.MEMSET

    overloaded = True

    @declared_attr
    def pointer(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 0)

    @declared_attr
    def value(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 1)

    @declared_attr
    def size(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 2)

    @declared_attr
    def volatile(cls) -> RelationshipProperty:
        return cls.operand_relationship(cls.cpg, 3)


class UnclassifiedNode(NodeMixin):
    _kind = NodeKind.UNCLASSIFIED_NODE


class ParamBinding(NodeMixin):
    _kind = NodeKind.PARAM_BINDING

    @declared_attr
    def callsite(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.CALL_TO_PARAM_BINDING, cls.cpg.Instruction, backref="binds"
        )

    @declared_attr
    def operand(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.OPERAND_TO_PARAM_BINDING, cls.cpg.Instruction, backref="bound_by"
        )

    @declared_attr
    def argument(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.PARAM_BINDING_TO_ARG, cls.cpg.Argument, backref="bound_by"
        )


class CallReturn(NodeMixin):
    _kind = NodeKind.CALL_RETURN

    @declared_attr
    def callsite(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.CALL_RETURN_TO_CALLER, cls.cpg.Instruction, backref="returns_from"
        )

    @declared_attr
    def bindings(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.SAME_CALL, cls.cpg.ParamBinding, backref="return_")

    @declared_attr
    def instruction(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN,
            cls.cpg.Instruction,
            backref="returns_to",
        )

    @declared_attr
    def value(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.RETURN_VALUE_TO_CALL_RETURN,
            cls.cpg.Instruction,
            backref="returned_by",
        )

    @declared_attr
    def returns_from(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.RETURN_INSTRUCTION_TO_CALL_RETURN, cls.cpg.Instruction
        )


class LocalVariable(Variable):
    _kind = NodeKind.LOCAL_VARIABLE

    @declared_attr
    def parent_function(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.FUNCTION_TO_LOCAL_VARIABLE,
            cls.cpg.Function,
            backref="local_variables",
        )

    @declared_attr
    def llvm_type(cls) -> RelationshipProperty:
        return cls.edge_relationship(EdgeKind.HAS_LLVM_TYPE, cls.cpg.LLVMType)

    @declared_attr
    def as_dwarf(cls) -> RelationshipProperty:
        return cls.edge_relationship(
            EdgeKind.LOCAL_VARIABLE_TO_DWARF_LOCAL_VARIABLE,
            cls.cpg.DWARFLocalVariable,
            backref="as_llvm",
        )

    @property
    def memory_locations(cls) -> Optional[List[MemoryLocation]]:
        if not cls.allocation_site:
            return None
        return cls.allocation_site.allocates

    @declared_attr
    def allocation_site(cls) -> RelationshipProperty:
        return relationship(
            cls.cpg.Alloca,
            secondary=cls.cpg.Edge.__table__,
            primaryjoin=and_(
                cls.cpg.Edge.target == cls.uuid,
                cls.cpg.Edge.kind == EdgeKind.CREATES_VAR,
            ),
            secondaryjoin=cls.cpg.Alloca.uuid == cls.cpg.Edge.source,
            uselist=False,
            doc="Link to the Alloca that created this LocalVariable",
        )



"""Module implementing an enhanced string column type for SQLAlchemy with a support for regular
expression operators in Postgres and SQLite.

http://xion.io/post/code/sqlalchemy-regex-filters.html
https://gist.github.com/Xion/204ddbd020f1a4275a53
"""
from __future__ import annotations

from re import Pattern
from typing import Any

from sqlalchemy.ext.hybrid import Comparator
from sqlalchemy.sql.expression import BinaryExpression, literal
from sqlalchemy.sql.operators import custom_op

__all__ = ["StringComparator"]


class StringComparator(Comparator):
    """A custom comparator for strings supporting a ``re_match`` operation."""

    def re_match(self, other: Pattern) -> _RegexMatchExpression:
        other0: Any = other
        if isinstance(other, Pattern):
            other0 = other.pattern
        return _RegexMatchExpression(self.__clause_element__(), literal(other0), custom_op("~"))

    def operate(self, op: Any, other: Any, **_kwargs: Any) -> Any:
        return op(self.__clause_element__(), other)


class _RegexMatchExpression(BinaryExpression):
    """Represents matching of a column againsts a regular expression."""



"""Nodes in the x86_64 binary AST.

At runtime, the models here are accessed via attributes on a CPG, not directly.
"""

from __future__ import annotations

from typing import List

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import RelationshipProperty

from mate_common.models.cpg_types import EdgeKind, NodeKind, UsedMemory, UsedRegister
from mate_query.cpg.models.node._typechecking import NodeMixin


class ASMBlock(NodeMixin):
    _kind = NodeKind.ASM_BLOCK

    @declared_attr
    def entry(cls) -> RelationshipProperty:
        """Returns the entry `ASMInst` for this block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_ENTRY_INSTRUCTION, cls.cpg.ASMInst)

    @declared_attr
    def terminator(cls) -> RelationshipProperty:
        """Returns the terminating `ASMInst` for this block."""
        return cls.edge_relationship(EdgeKind.BLOCK_TO_TERMINATOR_INSTRUCTION, cls.cpg.ASMInst)

    @declared_attr
    def instructions(cls) -> RelationshipProperty:
        """Returns a sequence of every `ASMInst` in this block."""
        return cls.edge_relationship(
            EdgeKind.INSTRUCTION_TO_PARENT_BLOCK, cls.cpg.ASMInst, backref="asm_block"
        )


class ASMInst(NodeMixin):
    _kind = NodeKind.ASM_INST

    @property
    def used_registers(self) -> List[UsedRegister]:
        """Returns a sequence of `UsedRegister` for each register used by this instruction."""
        return [UsedRegister(**r) for r in self.attributes["used_registers"]]

    @property
    def used_memory(self) -> List[UsedMemory]:
        """Returns a sequence of `UsedMemory` for each discrete memory access in this
        instruction."""
        return [UsedMemory(**r) for r in self.attributes["used_memory"]]


class ASMGlobalVariable(NodeMixin):
    _kind = NodeKind.ASM_GLOBAL_VARIABLE

    @declared_attr
    def dwarf_type(cls) -> RelationshipProperty:
        """Returns the `DWARFType` for this global variable."""
        return cls.edge_relationship(
            EdgeKind.HAS_DWARF_TYPE, cls.cpg.DWARFType, backref="global_variables"
        )


class PLTStub(NodeMixin):
    _kind = NodeKind.PLT_STUB

    @declared_attr
    def vtables(cls) -> RelationshipProperty:
        """Returns a sequence of `VTable` instances associated with this PLT stub."""
        return cls.edge_relationship(
            EdgeKind.PLT_STUB_TO_VTABLE,
            cls.cpg.VTable,
            backref="plt_stubs",
        )


class VTable(NodeMixin):
    _kind = NodeKind.VTABLE

    @hybrid_property
    def members(self):
        """The members of this virtual table."""
        """Returns the individual members of this virtual table."""
        return self.attributes["members"]

    @members.expression  # type: ignore[no-redef]
    def members(cls):
        return cls.attributes["members"].as_list()
